/* Compact floating-point encoding, procedures for timing test program.
   Written by Radford M. Neal, 2015. */

#include "cflt.h"


static double pow10[16] = 
{ 1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7,
  1e8, 1e9, 1e10, 1e11, 1e12, 1e13, 1e14, 1e15 
};


/* COPY TO UNCOMPRESSED ARRAY. */

void cflt_copy_uncompressed (int n, double *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = v[i];
  }
}

void cflt_copy_compressed (int n, cflt_t *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = cflt_decode(v[i]);
  }
}

void cflt_copy_padded_compressed (int n, uint64_t *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = cflt_decode((cflt_t)v[i]);
  }
}

void cflt_copy_scaled_float_div (int n, float *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = ((double)v[i] / 1000.0);
  }
}

void cflt_copy_scaled_integer_div (int n, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = ((double)v[i] / 1000.0);
  }
}

void cflt_copy_decimal_float (int n, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = ((double)(v[i] >> 4) / pow10[v[i]&0xf]);
  }
}

void cflt_copy_scaled_integer_mul (int n, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = ((double)v[i] * 0.001);
  }
}


/* SUM AN ARRAY OF VALUES STORED IN VARIOUS WAYS. */

double cflt_sum_uncompressed (int n, double *v)
{
  double s = 0.0;
  int i;
  for (i = 0; i < n; i++) 
  { s = s + v[i];
  }
  return s;
}

double cflt_sum_compressed (int n, cflt_t *v)
{
  double s = 0.0;
  int i;
  for (i = 0; i < n; i++) 
  { s = s + cflt_decode(v[i]);
  }
  return s;
}

double cflt_sum_padded_compressed (int n, uint64_t *v)
{
  double s = 0.0;
  int i;
  for (i = 0; i < n; i++) 
  { s = s + cflt_decode((cflt_t)v[i]);
  }
  return s;
}

double cflt_sum_scaled_float_div (int n, float *v)
{
  double s = 0.0;
  int i;
  for (i = 0; i < n; i++) 
  { s = s + (double)v[i] / 1000.0;
  }
  return s;
}

double cflt_sum_scaled_integer_div (int n, int *v)
{
  double s = 0.0;
  int i;
  for (i = 0; i < n; i++) 
  { s = s + (double)v[i] / 1000.0;
  }
  return s;
}

double cflt_sum_decimal_float (int n, int *v)
{
  double s = 0.0;
  int i;
  for (i = 0; i < n; i++) 
  { s = s + (double)(v[i] >> 4) / pow10[v[i]&0xf];
  }
  return s;
}

/* this one doesn't work, since *0.001 is not the same as /1000. */
double cflt_sum_scaled_integer_mul (int n, int *v)
{
  double s = 0.0;
  int i;
  for (i = 0; i < n; i++) 
  { s = s + (double)v[i] * 0.001;
  }
  return s;
}


/* MULTIPLY AN ARRAY OF VALUES STORED IN VARIOUS WAYS BY A SCALAR. 
   Store the result uncompressed.  */

void cflt_mul_uncompressed (int n, double a, double *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = a * v[i];
  }
}

void cflt_mul_compressed (int n, double a, cflt_t *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = a * cflt_decode(v[i]);
  }
}

void cflt_mul_padded_compressed (int n, double a, uint64_t *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = a * cflt_decode((cflt_t)v[i]);
  }
}

void cflt_mul_scaled_float_div (int n, double a, float *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = a * ((double)v[i] / 1000.0);
  }
}

void cflt_mul_scaled_integer_div (int n, double a, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = a * ((double)v[i] / 1000.0);
  }
}

void cflt_mul_decimal_float (int n, double a, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = a * ((double)(v[i] >> 4) / pow10[v[i]&0xf]);
  }
}

/* this one doesn't work, since *0.001 is not the same as /1000. */
void cflt_mul_scaled_integer_mul (int n, double a, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = a * ((double)v[i] * 0.001);
  }
}


/* ADD TWO ARRAYS OF VALUES STORED IN VARIOUS WAYS.  Result is stored
   uncompressed. */

void cflt_add_uncompressed (int n, double *u, double *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = u[i] + v[i];
  }
}

void cflt_add_compressed (int n, cflt_t *u, cflt_t *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = cflt_decode(u[i]) + cflt_decode(v[i]);
  }
}

void cflt_add_padded_compressed (int n, uint64_t *u, uint64_t *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = cflt_decode((cflt_t)u[i]) + cflt_decode((cflt_t)v[i]);
  }
}

void cflt_add_scaled_float_div (int n, float *u, float *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = (double)u[i] / 1000.0 + (double)v[i] / 1000.0;
  }
}

void cflt_add_scaled_integer_div (int n, int *u, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = (double)u[i] / 1000.0 + (double)v[i] / 1000.0;
  }
}

void cflt_add_decimal_float (int n, int *u, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = (double)(u[i] >> 4) / pow10[u[i]&0xf]
            + (double)(v[i] >> 4) / pow10[v[i]&0xf];
  }
}

/* this one doesn't work, since *0.001 is not the same as /1000. */
void cflt_add_scaled_integer_mul (int n, int *u, int *v, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = (double)u[i] * 0.001 + (double)v[i] * 0.001;
  }
}


/* FIND LINEAR COMBINATION OF THREE ARRAYS STORED IN VARIOUS WAYS.  Result is 
   stored uncompressed. */

void cflt_lcmb_uncompressed (int n, double *u, double *v, double *w, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = 1.1*u[i] + 2.2*v[i] + 3.3*w[i];
  }
}

void cflt_lcmb_compressed (int n, cflt_t *u, cflt_t *v, cflt_t *w, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = 1.1*cflt_decode(u[i]) + 2.2*cflt_decode(v[i]) 
                                 + 3.3*cflt_decode(w[i]);
  }
}

void cflt_lcmb_padded_compressed (int n, uint64_t *u, uint64_t *v, uint64_t *w,
                                  double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = 1.1*cflt_decode((cflt_t)u[i]) + 2.2*cflt_decode((cflt_t)v[i])
                                         + 3.3*cflt_decode((cflt_t)w[i]);
  }
}

void cflt_lcmb_scaled_float_div (int n, float *u, float *v, float *w, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = 1.1*((double)u[i] / 1000.0) + 2.2*((double)v[i] / 1000.0)
                                       + 3.3*((double)w[i] / 1000.0);
  }
}

void cflt_lcmb_scaled_integer_div (int n, int *u, int *v, int *w, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = 1.1*((double)u[i] / 1000.0) + 2.2*((double)v[i] / 1000.0) 
                                       + 3.3*((double)w[i] / 1000.0);
  }
}

void cflt_lcmb_decimal_float (int n, int *u, int *v, int *w, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = 1.1*((double)(u[i] >> 4) / pow10[u[i]&0xf])
            + 2.2*((double)(v[i] >> 4) / pow10[v[i]&0xf])
            + 3.3*((double)(w[i] >> 4) / pow10[w[i]&0xf]);
  }
}

/* this one doesn't work, since *0.001 is not the same as /1000. */
void cflt_lcmb_scaled_integer_mul (int n, int *u, int *v, int *w, double *r)
{
  int i;
  for (i = 0; i < n; i++) 
  { r[i] = 1.1*((double)u[i] * 0.001) + 2.2*((double)v[i] * 0.001)
                                      + 3.3*((double)w[i] * 0.001);
  }
}
