/* Compact floating-point encoding, timing test program.
   Written by Radford M. Neal, 2015. */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#include "cflt.h"

/* Some of these are no longer used... */

void cflt_copy_uncompressed (int, double *, double *);
void cflt_copy_compressed (int, cflt_t *, double *);
void cflt_copy_padded_compressed (int, uint64_t *, double *);
void cflt_copy_scaled_float_div (int, float *, double *);
void cflt_copy_scaled_integer_div (int, int *, double *);
void cflt_copy_decimal_float (int, int *, double *);
void cflt_copy_scaled_integer_mul (int, int *, double *);

double cflt_sum_uncompressed (int, double *);
double cflt_sum_compressed (int, cflt_t *);
double cflt_sum_padded_compressed (int, uint64_t *);
double cflt_sum_scaled_float_div(int, float *);
double cflt_sum_scaled_integer_div(int, int *);
double cflt_sum_decimal_float(int, int *);
double cflt_sum_scaled_integer_mul(int, int *);

void cflt_mul_uncompressed (int, double, double *, double *);
void cflt_mul_compressed (int, double, cflt_t *, double *);
void cflt_mul_padded_compressed (int, double, uint64_t *, double *);
void cflt_mul_scaled_float_div (int, double, float *, double *);
void cflt_mul_scaled_integer_div (int, double, int *, double *);
void cflt_mul_decimal_float (int, double, int *, double *);
void cflt_mul_scaled_integer_mul (int, double, int *, double *);

void cflt_add_uncompressed (int, double *, double *, double *);
void cflt_add_compressed (int, cflt_t *, cflt_t *, double *);
void cflt_add_padded_compressed (int, uint64_t *, uint64_t *, double *);
void cflt_add_scaled_float_div (int, float *, float *, double *);
void cflt_add_scaled_integer_div (int, int *, int *, double *);
void cflt_add_decimal_float (int, int *, int *, double *);
void cflt_add_scaled_integer_mul (int, int *, int *, double *);

void cflt_lcmb_uncompressed (int, double *, double *, double *, double *);
void cflt_lcmb_compressed (int, cflt_t *, cflt_t *, cflt_t *, double *);
void cflt_lcmb_padded_compressed (int, uint64_t *, uint64_t *, uint64_t *, 
                                  double *);
void cflt_lcmb_scaled_float_div (int, float *, float *, float *, double *);
void cflt_lcmb_scaled_integer_div (int, int *, int *, int *, double *);
void cflt_lcmb_decimal_float (int, int *, int *, int *, double *);
void cflt_lcmb_scaled_integer_mul (int, int *, int *, int *, double *);

int cflt_args (char **);

#ifdef CLOCK_REALTIME

#define START_CLOCK \
  ( start_clock = clock(), clock_gettime(CLOCK_REALTIME,&start_time) )
#define END_CLOCK \
  ((double) (clock() - start_clock) / CLOCKS_PER_SEC)
#define END_CLOCK2 \
  ( clock_gettime(CLOCK_REALTIME,&end_time), \
    (end_time.tv_sec+end_time.tv_nsec*1e-9) \
      - (start_time.tv_sec+start_time.tv_nsec*1e-9) )

#else

#define START_CLOCK ( start_clock = clock() )
#define END_CLOCK ((double) (clock() - start_clock) / CLOCKS_PER_SEC)
#define END_CLOCK2 END_CLOCK

#endif

#define CHECK_DST \
  do { \
    for (i = 0; i < N; i++) \
    { if (dst0[i] != dst[i]) \
      { printf("  DIFFERENT RESULT!\n\n"); \
        break; \
      } \
    } \
  } while (0)


int main (int argc, char **argv)
{
  double r_unc, r_cmp, r_sclm, r_scld, r_sclf, r_dflt, a;
  clock_t start_clock; struct timespec start_time, end_time;
  double *flt, *flt2, *flt3, *dst0, *dst;
  cflt_t *cflt, *cflt2, *cflt3;
  uint64_t *pcflt, *pcflt2, *pcflt3;
  int32_t *scli, *scli2, *scli3, *dflt, *dflt2, *dflt3;
  float *sing, *sing2, *sing3;
  int64_t rnd;
  char type;
  int N, R;
  int i, j;

  if (argc < 4 || strcmp(argv[1],"f") != 0 && strcmp(argv[1],"v") != 0
       || cflt_args(argv+3) != 0)
  {
#   if CFLT_2_MASKS | CFLT_RECODE
      fprintf(stderr,
       "Usage:  cflt[2345]-time type N R bits bits2 epos { NA | n/m | n.m }\n");
#   else
      fprintf(stderr,
       "Usage:  cflt1-time type N R bits { NA | n/m | n.m }\n");
#   endif
    exit(1);
  }

  type = *argv[1];
  N = atoi(argv[2]);
  R = atoi(argv[3]);

  flt = calloc ((size_t)N, sizeof *flt);
  if (flt == 0) abort();
  flt2 = calloc ((size_t)N, sizeof *flt2);
  if (flt2 == 0) abort();
  flt3 = calloc ((size_t)N, sizeof *flt3);
  if (flt3 == 0) abort();
  dst0 = calloc ((size_t)N, sizeof *dst0);
  if (dst0 == 0) abort();
  dst = calloc ((size_t)N, sizeof *dst);
  if (dst == 0) abort();
  cflt = calloc ((size_t)N, sizeof *cflt);
  if (cflt == 0) abort();
  cflt2 = calloc ((size_t)N, sizeof *cflt2);
  if (cflt2 == 0) abort();
  cflt3 = calloc ((size_t)N, sizeof *cflt3);
  if (cflt3 == 0) abort();
  pcflt = calloc ((size_t)N, sizeof *pcflt);
  if (pcflt == 0) abort();
  pcflt2 = calloc ((size_t)N, sizeof *pcflt2);
  if (pcflt2 == 0) abort();
  pcflt3 = calloc ((size_t)N, sizeof *pcflt3);
  if (pcflt3 == 0) abort();
  scli = calloc ((size_t)N, sizeof *scli);
  if (scli == 0) abort();
  scli2 = calloc ((size_t)N, sizeof *scli2);
  if (scli2 == 0) abort();
  scli3 = calloc ((size_t)N, sizeof *scli3);
  if (scli3 == 0) abort();
  dflt = calloc ((size_t)N, sizeof *dflt);
  if (dflt == 0) abort();
  dflt2 = calloc ((size_t)N, sizeof *dflt2);
  if (dflt2 == 0) abort();
  dflt3 = calloc ((size_t)N, sizeof *dflt3);
  if (dflt3 == 0) abort();
  sing = calloc ((size_t)N, sizeof *sing);
  if (sing == 0) abort();
  sing2 = calloc ((size_t)N, sizeof *sing2);
  if (sing2 == 0) abort();
  sing3 = calloc ((size_t)N, sizeof *sing3);
  if (sing3 == 0) abort();

  rnd = 2345678;

  for (j = 0; j < 3*N; j++)
  { 
    int dive =  type == 'f' ? 3 : j % 3 == 0 ? 2 : j % 3 == 1 ? 3 : 4;
    double div = dive == 2 ? 100.0 : dive == 3 ? 1000.0 : 10000.0;
    double fl;
    cflt_t cf;
    int sc;

    rnd = (rnd*8192) % 67101323; /* from "Multiplicative congruential pseudo-
                                    random number generators", Downham&Roberts*/

    if (rnd == 0) abort();

    sc = (int) (1999999 * ((double)rnd / 67101323)) - 999999;
    fl = (double)sc / div;
    if (!cflt_encode(fl,&cf))
    { printf ("Can't encode %f\n",fl);
      exit(1);
    }

    if (j < N) 
    { i = j;
      scli[i] = sc;
      dflt[i] = (sc << 4) + dive;
      sing[i] = (float)sc;
      flt[i] = fl;
      cflt[i] = cf;
      pcflt[i] = cf;
    }
    else if (j < 2*N)
    { i = j - N;
      scli2[i] = sc;
      dflt2[i] = (sc << 4) + dive;
      sing2[i] = (float)sc;
      flt2[i] = fl;
      cflt2[i] = cf;
      pcflt2[i] = cf;
    }
    else
    { i = j - 2*N;
      scli3[i] = sc;
      dflt3[i] = (sc << 4) + dive;
      sing3[i] = (float)sc;
      flt3[i] = fl;
      cflt3[i] = cf;
      pcflt3[i] = cf;
    }
  }

  printf("Type %c,  N = %d,  R = %d\n\n",type,N,R);

  printf("\nMAKING UNCOMPRESSED COPY\n\n");

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_copy_uncompressed(N,flt,dst);
  printf ("Uncompressed:         %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);

  for (i = 0; i < N; i++) dst0[i] = dst[i];

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_copy_compressed(N,cflt,dst);
  printf ("Compressed:           %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);
  CHECK_DST;

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_copy_decimal_float(N,dflt,dst);
  printf ("Decimal float:        %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);
  CHECK_DST;

  printf("\nSUMMING AN ARRAY\n\n");

  START_CLOCK;
  for (i = 0; i < R; i++) r_unc = cflt_sum_uncompressed(N,flt);
  printf("Uncompressed sum:         %.2f, %.3f/%.3f seconds\n\n", r_unc, 
          END_CLOCK, END_CLOCK2);

  START_CLOCK;
  for (i = 0; i < R; i++) r_cmp = cflt_sum_compressed(N,cflt);
  printf("Compressed sum:           %.2f, %.3f/%.3f seconds\n\n", r_cmp, 
          END_CLOCK, END_CLOCK2);
  if (r_unc != r_cmp) printf("  DIFFERENT RESULT!\n\n");

  START_CLOCK;
  for (i = 0; i < R; i++) r_dflt = cflt_sum_decimal_float(N,dflt);
  printf("Decimal float sum:        %.2f, %.3f/%.3f seconds\n\n", r_dflt,
          END_CLOCK, END_CLOCK2);
  if (r_unc != r_dflt) printf("  DIFFERENT RESULT!\n\n");

  printf("\nMULTIPLYING AN ARRAY BY A SCALAR\n\n");

  a = 123.456789;

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_mul_uncompressed(N,a,flt,dst);
  printf ("Uncompressed:         %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);

  for (i = 0; i < N; i++) dst0[i] = dst[i];

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_mul_compressed(N,a,cflt,dst);
  printf ("Compressed:           %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);
  CHECK_DST;

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_mul_decimal_float(N,a,dflt,dst);
  printf ("Decimal float:        %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);
  CHECK_DST;
 
  printf("\nADDING TWO ARRAYS\n\n");

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_add_uncompressed(N,flt,flt2,dst);
  printf ("Uncompressed:         %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);

  for (i = 0; i < N; i++) dst0[i] = dst[i];

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_add_compressed(N,cflt,cflt2,dst);
  printf ("Compressed:           %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);
  CHECK_DST;

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_add_decimal_float(N,dflt,dflt2,dst);
  printf ("Decimal float:        %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);
  CHECK_DST;
 
  printf("\nLINEARLY COMBINING THREE ARRAYS\n\n");

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_lcmb_uncompressed(N,flt,flt2,flt3,dst);
  printf ("Uncompressed:         %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);

  for (i = 0; i < N; i++) dst0[i] = dst[i];

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_lcmb_compressed(N,cflt,cflt2,cflt3,dst);
  printf ("Compressed:           %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);
  CHECK_DST;

  START_CLOCK;
  for (i = 0; i < R; i++) cflt_lcmb_decimal_float(N,dflt,dflt2,dflt3,dst);
  printf ("Decimal float:        %.3f/%.3f seconds\n\n", END_CLOCK, END_CLOCK2);
  CHECK_DST;

  return 0;
}
