/* Compact floating-point encoding, function and data structure declarations.
   Written by Radford M. Neal, 2015. */

#include <stdio.h>

#include "cflt.h"         /* Includes definitions of inline functions */


/* DEFINITIONS OF GLOBALS.  Declared in cflt.h. */

uint32_t *cflt_table;     /* The lookup table for filling in the low-order
                             32 bits of the mantissa */

int *cflt_used;           /* Flags for whether a table entry has already
                             been put to use */

int cflt_bits;            /* Number of bits of mantissa to look at */
int cflt_bits2;           /* Number of bits of exponent to look at */
int cflt_size;            /* Size of (uncondensed) value table */
unsigned cflt_mask;       /* Mask for bits used to index table */
unsigned cflt_mask2;      /* Second mask for bits used to index table */
int cflt_shift2;          /* Amount to shift before using mask2 */

int cflt_ncodes;          /* Number of codes used in recoding */
cflt_recode_t *cflt_rtable; /* Recoding table (NULL if none) */
uint16_t *cflt_itable;    /* Indirect indexing table (NULL if none) */


/* START DEFINING A DECODING TABLE.  Initializes the decoding table,
   using a specified number of bits from the low-order end of the
   high-order part of the mantissa, plus a specified number of bits
   from part of the exponent for the 2 masks variant, to look up the
   remaining part of the mantissa.  Initially, the lookup table
   contains all zeros, so the set of representable values consists of
   those with zeros in the the low-order 32 bits of the mantissa. */

# if CFLT_2_MASKS | CFLT_RECODE
void cflt_start (int bits, int bits2, int epos) {
# else
void cflt_start (int bits) { int bits2 = 0, epos = 0;
# endif

  int i;

  cflt_bits = bits;
  cflt_bits2 = bits2;
  cflt_size = 1 << (bits + bits2);
  cflt_mask = (1 << bits) - 1;
  cflt_mask2 = ((1 << bits2) - 1) << bits;
  cflt_shift2 = 20 + epos - bits;

  if (cflt_table != 0) free(cflt_table);
  cflt_table = calloc ((size_t) cflt_size, sizeof *cflt_table);
  if (cflt_table == 0) abort();
  
  if (cflt_used != 0) free(cflt_used);
  cflt_used = calloc ((size_t) cflt_size, sizeof *cflt_used);
  if (cflt_used == 0) abort();

  for (i = 0; i < cflt_size; i++) 
  { cflt_used[i] = 0;
    cflt_table[i] = 0;
  }

# if CFLT_RECODE
    if (cflt_rtable != 0) free(cflt_rtable);
    cflt_rtable = calloc ((size_t) (1 << cflt_bits), sizeof *cflt_rtable);
    if (cflt_rtable == 0) abort();
# endif

# if CFLT_INDIRECT
    if (cflt_itable != 0) free(cflt_itable);
    cflt_itable = calloc ((size_t) cflt_size, sizeof *cflt_itable);
    if (cflt_itable == 0) abort();
# endif
}


/* SET AN ENTRY IN THE LOOKUP TABLE TO REPRESENT A VALUE.  Sets the
   required entry provided it has not previously been set (in which
   case 0 is returned, with 1 returned for success).  This function
   should be called for as many values as desired to initialize the
   table, before any values are encoded. */

int cflt_use_value (double value)
{
# if CFLT_TYPE_PUN == -2
    uint64_t u;
    *(double*)&u = value;
    return cflt_use_parts ((uint32_t)(u>>32), (uint32_t)(u&0xffffffff));
# elif CFLT_TYPE_PUN == -1
    union { double d; uint64_t v; } u;
    u.d = value;
    return cflt_use_parts (u.v >> 32, u.v & 0xffffffff);
# else
    union { double d; uint32_t w[2]; } u;
    u.d = value;
    return cflt_use_parts (u.w[CFLT_TYPE_PUN], u.w[1-CFLT_TYPE_PUN]);
# endif
}


/* SET AN ENTRY GIVEN TWO 32-BIT PARTS OF A DOUBLE VALUE.  Like
   cflt_use_value, except that the value is specified by its high
   and low 32-bit unsigned int parts. */

int cflt_use_parts (uint32_t high, uint32_t low)
{
  int ix;
# if CFLT_2_MASKS | CFLT_RECODE
    ix = (high & cflt_mask) + ((high >> cflt_shift2) & cflt_mask2);
# else
    ix = high & cflt_mask;
# endif
  if (cflt_used[ix])  
  { return cflt_table[ix] == low;
  }
  else
  { cflt_used[ix] = 1;
    cflt_table[ix] = low;
    return 1;
  }
}


/* COMPUTE RECODING TABLE.  Does nothing if CFLT_RECODE is undefined or 
   zero, otherwise sets contents of cflt_rtable and updates cflt_table 
   and cflt_size. */

static uint32_t *tmp_table;
static int *tmp_used;
static int n1, n2;

static inline int merge (cflt_t *at, int *au, cflt_t *bt, int *bu, int *ex)
{ 
  int h, j, x, besth, bestx, full;

  bestx = -1;
  besth = -1;

  for (h = 0; h < n2 && bestx != 0; h++)
  { x = 0;
    for (j = 0; j < n2; j++)
    { if (au[j])
      { if (bu[j^h])
        { if (at[j] != bt[j^h])
          { break;
          }
        }
        else
        { x += 1;
        }
      }
    }
    if (j == n2 && (besth == -1 || x < bestx))
    { bestx = x;
      besth = h;
    }
  }

  if (besth == -1)
  { return 0;
  }

  if (ex) 
  { *ex = bestx;
    return 1;
  }

  for (j = 0; j < n2; j++)
  { if (au[j] && !bu[j^besth])
    { bt[j^besth] = at[j];
      bu[j^besth] = 1;
    }
  }

  return 1;
}

static int cmp_nused (const void *a, const void *b)
{
  const int *ai = a, *bi = b;
  int na = 0, nb = 0;
  int j;
  for (j = 0; j < n2; j++) 
  { na += tmp_used [(*ai << cflt_bits2) + j];
    nb += tmp_used [(*bi << cflt_bits2) + j];
  }
  return nb>na ? 1 : nb<na ? -1 : *ai>*bi ? 1 : *ai<*bi ? -1 : 0;
}

void cflt_recode (void)
{
  int h, i, j, k, m, ix, iy, ex;
  int mrg_cnt, nz_cnt;
  int *tmp_reorder;
  uint32_t *tmp2_table;
  int *tmp2_used;

  printf("Recoding...\n");

  n1 = 1 << cflt_bits;
  n2 = 1 << cflt_bits2;

  /* Allocate temporary tables, allowing for an extra "zero" entry. */

  tmp_table = calloc ((size_t) cflt_size+n2, sizeof *tmp_table);
  if (tmp_table == 0) abort();
  tmp_used = calloc ((size_t) cflt_size+n2, sizeof *tmp_used);
  if (tmp_used == 0) abort();

  for (i = 0; i < cflt_size; i++) 
  { tmp_table[ ((i&cflt_mask)<<cflt_bits2) + (i>>cflt_bits)] = cflt_table[i];
    tmp_used [ ((i&cflt_mask)<<cflt_bits2) + (i>>cflt_bits)] = cflt_used[i];
  }

  /* Eliminate unused entries. */

  cflt_ncodes = n1;
  mrg_cnt = 0;

# if 0
    printf("Before eliminating unused entries:\n");
    for (i = 0; i < cflt_ncodes; i++)
    { for (j = 0; j < n2; j++)
      { printf(" %d:%08x",
         tmp_used[(i<<cflt_bits2)+j], tmp_table[(i<<cflt_bits2)+j]);
      }
      printf("\n");
    }
    printf("\n");
# endif

  for (i = cflt_ncodes-1; i >= 0; i--)
  {
    for (j = 0; j < n2; j++)
    { if (tmp_used [ (i << cflt_bits2) + j ])
      { break;
      }
    }

    if (j == n2)
    { for (k = (i << cflt_bits2) + n2; k < cflt_ncodes << cflt_bits2; k++)
      { tmp_table[k-n2] = tmp_table[k];
        tmp_used[k-n2] = tmp_used[k];
      }
      mrg_cnt += 1;
      cflt_ncodes -= 1;
    }
  }

  printf("  Eliminated %d unused entries\n",mrg_cnt);

# if 0
    printf("After eliminating unused entries:\n");
    for (i = 0; i < cflt_ncodes; i++)
    { for (j = 0; j < n2; j++)
      { printf(" %d:%08x",
         tmp_used[(i<<cflt_bits2)+j], tmp_table[(i<<cflt_bits2)+j]);
      }
      printf("\n");
    }
    printf("\n");
# endif

  /* Sort in decreasing order by number of used entries. */

  tmp_reorder = calloc ((size_t) cflt_ncodes, sizeof *tmp_reorder);
  if (tmp_reorder == 0) abort();
  tmp2_table = calloc ((size_t) cflt_size+n2, sizeof *tmp2_table);
  if (tmp2_table == 0) abort();
  tmp2_used = calloc ((size_t) cflt_size+n2, sizeof *tmp2_used);
  if (tmp2_used == 0) abort();

  for (i = 0; i < cflt_ncodes; i++) tmp_reorder[i] = i;

  qsort (tmp_reorder, (size_t) cflt_ncodes, sizeof *tmp_reorder, cmp_nused);

  for (i = 0; i < cflt_ncodes; i++) 
  { for (j = 0; j < n2; j++)
    { tmp2_used [ (i << cflt_bits2) + j ] 
       = tmp_used [ (tmp_reorder[i] << cflt_bits2) + j ];
      tmp2_table [ (i << cflt_bits2) + j ] 
       = tmp_table [ (tmp_reorder[i] << cflt_bits2) + j ];
    }
  }

  free (tmp_table);
  tmp_table = tmp2_table;
  free (tmp_used);
  tmp_used = tmp2_used;

# if 0
    printf("After sorting:\n");
    for (i = 0; i < cflt_ncodes; i++)
    { for (j = 0; j < n2; j++)
      { printf(" %d:%08x",
         tmp_used[(i<<cflt_bits2)+j], tmp_table[(i<<cflt_bits2)+j]);
      }
      printf("\n");
    }
    printf("\n");
# endif

  /* Do costless merges. */

  mrg_cnt = 0;

  for (i = 0; i < cflt_ncodes; i++)
  { k = i + 1;
    while (k < cflt_ncodes)
    { if (merge (tmp_table + (k<<cflt_bits2), tmp_used  + (k<<cflt_bits2),
                 tmp_table + (i<<cflt_bits2), tmp_used  + (i<<cflt_bits2), &ex)
          && ex==0)
      { merge (tmp_table + (k<<cflt_bits2), tmp_used  + (k<<cflt_bits2),
               tmp_table + (i<<cflt_bits2), tmp_used  + (i<<cflt_bits2), NULL);
        for (m = (k << cflt_bits2) + n2; m < cflt_ncodes << cflt_bits2; m++)
        { tmp_table[m-n2] = tmp_table[m];
          tmp_used[m-n2] = tmp_used[m];
        }
        mrg_cnt += 1;
        cflt_ncodes -= 1;
      }
      else
      { k += 1;
      }
    }
  }

  printf("  Merged %d entries costlessly\n",mrg_cnt);

# if 0
    printf("After costless merges:\n");
    for (i = 0; i < cflt_ncodes; i++)
    { for (j = 0; j < n2; j++)
      { printf(" %d:%08x",
         tmp_used[(i<<cflt_bits2)+j], tmp_table[(i<<cflt_bits2)+j]);
      }
      printf("\n");
    }
    printf("\n");
# endif

  /* Do merges for each code, prioritizing merging lots, then little extra. */

  mrg_cnt = 0;

  for (i = 0; i < cflt_ncodes; i++)
  { 
    for (;;)
    { int bestk = -1;
      int bestx = 1000000;
      int bestu = -1;
  
      for (k = i+1; k < cflt_ncodes; k++)
      { int nu = 0;
        for (j = 0; j < n2; j++)
        { nu += tmp_used [(k << cflt_bits2) + j];
        }
        if (nu < bestu) 
        { break;
        }
        if (merge (tmp_table + (k<<cflt_bits2), tmp_used  + (k<<cflt_bits2),
                 tmp_table + (i<<cflt_bits2), tmp_used  + (i<<cflt_bits2), &ex))
        { if (nu > bestu || nu == bestu && ex < bestx)
          { bestk = k;
            bestu = nu;
            bestx = ex;
          }
        }
      }
  
      if (bestk == -1) goto next_one;
  
      merge (tmp_table + (bestk<<cflt_bits2), tmp_used  + (bestk<<cflt_bits2),
             tmp_table + (i<<cflt_bits2), tmp_used  + (i<<cflt_bits2), NULL);
  
      for (m = (bestk << cflt_bits2) + n2; m < cflt_ncodes << cflt_bits2; m++)
      { tmp_table[m-n2] = tmp_table[m];
        tmp_used[m-n2] = tmp_used[m];
      }
      mrg_cnt += 1;
      cflt_ncodes -= 1;
    }
  next_one: ;
  }
  
  printf("  Merged %d entries\n",mrg_cnt);
  
# if 0
    printf("After merges:\n");
    for (i = 0; i < cflt_ncodes; i++)
    { for (j = 0; j < n2; j++)
      { printf(" %d:%08x",
         tmp_used[(i<<cflt_bits2)+j], tmp_table[(i<<cflt_bits2)+j]);
      }
      printf("\n");
    }
    printf("\n");
# endif

  /* Create the "zero" entry at the end. */

  for (j = 0; j < n2; j++)
  { tmp_table [(cflt_ncodes<<cflt_bits2)+j] = 0;
    tmp_used [(cflt_ncodes<<cflt_bits2)+j] = 0;
  }
  cflt_ncodes += 1;

# if 0
    printf("After all merges and zero entry created:\n");
    for (i = 0; i < cflt_ncodes; i++)
    { for (j = 0; j < n2; j++)
      { printf(" %d:%08x",
         tmp_used[(i<<cflt_bits2)+j], tmp_table[(i<<cflt_bits2)+j]);
      }
      printf("\n");
    }
    printf("\n");
# endif

  cflt_recode_t max = ~(cflt_recode_t)0;
  if (((uint32_t)cflt_ncodes << cflt_bits2) - 1 > max)
  { printf("Too many codes (%d) for recoding table type\n",cflt_ncodes);
    exit(1);
  }

  /* Loop to find a code for each value of the lower mantissa. */

  nz_cnt = 0;

  for (i = 0; i < n1; i++)
  { 
    /* Just use the "zero" entry if this mantissa value isn't used at all. */

    for (j = 0; j < n2; j++)
    { if (cflt_used [ i + (j << cflt_bits) ]) break;
    }

    if (j == n2)
    { cflt_rtable[i] = (cflt_ncodes-1) << cflt_bits2; /* code for all zeros */
      continue;
    }

    /* Look for a code entry, k, that will work with some value, h, 
       to XOR with the exponent part. */

    for (k = 0; ; k++)
    { if (k == cflt_ncodes-1) abort(); /* one should exist before 0 entry */
      for (h = 0; h < n2; h++)
      { int code = (k << cflt_bits2) + h;  /* a possible code to check */
        for (j = 0; j < n2; j++)  /* see if it works for all exponents */
        { iy = i + (j << cflt_bits);
          if (cflt_used[iy]) 
          { ix = code ^ j;
            if (!tmp_used[ix] || tmp_table[ix] != cflt_table[iy])
            { goto next_h;
            }
          }
        }
        cflt_rtable[i] = code;    /* it does work, so we're done */
        if (h != 0) nz_cnt += 1;
        goto next_i;
      next_h: ;
      }
    }
  next_i: ;
  }

  printf("Of %d mantissa values, %d have codes with non-zero lower part\n",
          n1, nz_cnt);

  /* Update things for using the recoded table. */

  cflt_shift2 += cflt_bits;
  cflt_mask2 >>= cflt_bits;
  cflt_size = cflt_ncodes << cflt_bits2;

  free(cflt_table);
  cflt_table = calloc ((size_t) cflt_size, sizeof *cflt_table);
  if (cflt_table == 0) abort();

  for (i = 0; i < cflt_size; i++) cflt_table[i] = tmp_table[i];
}



/* COMPUTE INDIRECT INDEXING TABLE AND CONDENSED VALUE TABLE.  Also returns
   the number of unique entries in the table.  If CFLT_INDIRECT is undefined
   or zero, that's all it does; otherwise it also sets up cflt_itable and
   condenses cflt_table. */

static int cflt_cmp (const void *a, const void *b) 
{ return *(cflt_t*)a > *(cflt_t*)b ? 1 : *(cflt_t*)a < *(cflt_t*)b ? -1 : 0;
}

int cflt_indirect (void)
{
  int i, j, unique;
  cflt_t *sorted_table;
  cflt_t *new_table;
  cflt_t val;

  sorted_table = calloc ((size_t) cflt_size, sizeof *sorted_table);
  if (sorted_table == 0) abort();

  for (i = 0; i < cflt_size; i++) sorted_table[i] = cflt_table[i];

  qsort (sorted_table, cflt_size, sizeof *sorted_table, cflt_cmp);

  unique = 1;
  for (i = 1; i < cflt_size; i++) 
  { if (sorted_table[i] != sorted_table[i-1])
    { sorted_table[unique] = sorted_table[i];
      unique += 1;
    }
  }

# if CFLT_INDIRECT

    if (unique > 0xffff) 
    { printf("Too many unique values (%d) for indirect table\n",unique);
      exit(1);
    }

    new_table = calloc ((size_t) unique, sizeof *cflt_table);
    if (new_table == 0) abort();

    for (j = 0; j < unique; j++) new_table[j] = sorted_table[j];

    for (i = 0; i < cflt_size; i++)
    { val = cflt_table[i];
      for (j = 0; new_table[j] != val; j++)
      { if (j == unique-1) abort();
      }
      cflt_itable[i] = j;
    }

    free(cflt_table);
    cflt_table = new_table;

# endif

  free (sorted_table);
  return unique;
}


/* RETURN THE NUMBER OF TABLE ENTRIES IN USE, AND MANTISSA/EXP VALUES USED.  
   Returns the total number of table entries used, the number of masked 
   mantissa values used, and the number of masked exponent values used.

   Note that the unused table entries, which are all zero, may still
   be used to encode values. */

void cflt_number_used (int *total_used, int *mantissa_used, int *exp_used)
{
  int i, j;

  *total_used = 0;
  for (i = 0; i < cflt_size; i++) 
  { *total_used += cflt_used[i];
  }

  *mantissa_used = 0;
  for (i = 0; i < (1<<cflt_bits); i++)
  { for (j = 0; j < (1<<cflt_bits2); j++)
    { if (cflt_used[(j << cflt_bits) + i])
      { *mantissa_used += 1;
        break;
      }
    }
  }

  *exp_used = 0;
  for (i = 0; i < (1<<cflt_bits2); i++)
  { for (j = 0; j < (1<<cflt_bits); j++)
    { if (cflt_used[(i << cflt_bits) + j])
      { *exp_used += 1;
        break;
      }
    }
  }
}
