/* Compact floating-point encoding, interface and inline functions.
   Written by Radford M. Neal, 2015. */

#include <stdlib.h>
#include <stdint.h>


/* DECLARATIONS. */

#ifndef CFLT_TYPE_PUN
#define CFLT_TYPE_PUN -1      /* -2 = cast double ptr to/from uint64_t ptr
                                 -1 = union with uint64_t, assumed same endian
                                  0 = union with two uint32_t, big endian 
                                  1 = union with two uint32_t, little endian */
#endif

typedef uint32_t cflt_t;      /* 32-bit compact floating-point data type */
typedef uint32_t cflt_recode_t; /* Type for recoding table entry */

extern uint32_t *cflt_table;  /* The lookup table for filling in the low-order
                                 32 bits of the mantissa */

extern int *cflt_used;        /* Flags for whether a table entry has already
                                 been put to use */

extern int cflt_bits;         /* Number of bits of mantissa to look at */
extern int cflt_bits2;        /* Number of bits of exponent to look at */
extern int cflt_size;         /* Size of (uncondensed) value table */
extern unsigned cflt_mask;    /* Mask for bits used to index table */
extern unsigned cflt_mask2;   /* Second mask for bits used to index table */
extern int cflt_shift2;       /* Amount to shift before using mask2 */

extern int cflt_ncodes;       /* Number of codes used in recoding */
extern cflt_recode_t *cflt_rtable; /* Recoding table (NULL if none) */
extern uint16_t *cflt_itable; /* Indirect indexing table (NULL if none) */


/* FUNCTIONS. */

# if CFLT_2_MASKS | CFLT_RECODE
void cflt_start (int,int,int);
# else
void cflt_start (int);
#endif
int cflt_use_value (double);
int cflt_use_parts (uint32_t high, uint32_t low);
void cflt_recode (void);
int cflt_indirect (void);
void cflt_number_used (int *, int *, int *);


/* EXTRACT A COMPRESSED ENCODING.  The cflt_encode function tries to
   encode a double value as a 32-bit unsigned int, returning 1 if
   sucessful, and 0 if not. */

static inline int cflt_encode (double value, cflt_t *encoding)
{ 
  cflt_t e;
  int ix;

# if CFLT_TYPE_PUN == -2
    uint64_t u;
    *(double*)&u = value;
    e = u >> 32;
#   if CFLT_RECODE
      ix = cflt_rtable [e & cflt_mask] ^ ((e >> cflt_shift2) & cflt_mask2);
#   elif CFLT_2_MASKS
      ix = (e & cflt_mask) + ((e >> cflt_shift2) & cflt_mask2);
#   else
      ix = e & cflt_mask;
#   endif
#   if CFLT_INDIRECT
      ix = cflt_itable[ix];
#   endif
    if ((u&0xffffffff) != cflt_table[ix]) return 0;
# elif CFLT_TYPE_PUN == -1
    union { double d; uint64_t v; } u;
    u.d = value;
    e = u.v >> 32;
#   if CFLT_RECODE
      ix = cflt_rtable [e & cflt_mask] ^ ((e >> cflt_shift2) & cflt_mask2);
#   elif CFLT_2_MASKS
      ix = (e & cflt_mask) + ((e >> cflt_shift2) & cflt_mask2);
#   else
      ix = e & cflt_mask;
#   endif
#   if CFLT_INDIRECT
      ix = cflt_itable[ix];
#   endif
    if ((u.v & 0xffffffff) != cflt_table[ix]) return 0;
# else
    union { double d; uint32_t w[2]; } u;
    u.d = value;
    e = u.w[CFLT_TYPE_PUN];
#   if CFLT_RECODE
      ix = cflt_rtable [e & cflt_mask] ^ ((e >> cflt_shift2) & cflt_mask2);
#   elif CFLT_2_MASKS
      ix = (e & cflt_mask) + ((e >> cflt_shift2) & cflt_mask2);
#   else
      ix = e & cflt_mask;
#   endif
#   if CFLT_INDIRECT
      ix = cflt_itable[ix];
#   endif
    if (u.w[1-CFLT_TYPE_PUN] != cflt_table[ix]) return 0;
# endif

  *encoding = e;
  return 1;
}


/* EXPAND A COMPRESSED ENCODING.  The cflt_decode function returns the
   double value decoded from a 32-bit compressed representation of a
   floating-point value. */

static inline double cflt_decode (cflt_t encoding)
{
  int ix;

# if CFLT_RECODE
    ix = cflt_rtable [encoding & cflt_mask] 
          ^ ((encoding >> cflt_shift2) & cflt_mask2);
# elif CFLT_2_MASKS
    ix = (encoding & cflt_mask) + ((encoding >> cflt_shift2) & cflt_mask2);
# else
    ix = encoding & cflt_mask;
# endif

# if CFLT_INDIRECT
    ix = cflt_itable[ix];
# endif

# if CFLT_TYPE_PUN == -2
    uint64_t u;
    u = ((uint64_t)encoding << 32) | cflt_table[ix];
    return *(double*)&u;
# elif CFLT_TYPE_PUN == -1
    union { double d; uint64_t v; } u;
    u.v = ((uint64_t)encoding << 32) | cflt_table[ix];
    return u.d;
# else
    union { double d; uint32_t w[2]; } u;
    u.w[CFLT_TYPE_PUN] = encoding;
    u.w[1-CFLT_TYPE_PUN] = cflt_table[ix];
    return u.d;
# endif
}
